/********************************************************************
 *		File			:		com.h
 *		Description		:		Component interface (MS-COM Compliant)
 *		Author			:		Jianwen Zhu
 *		Reference		:		http://www.microsoft.com/com
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#ifndef COM_H
#define COM_H

#include <stdio.h>
#include <config.h>

/* xxx: fix me -- Java like basic data type definitions */
typedef signed char				Jbyte;
typedef signed short int        Jshort;
typedef signed int              Jint;
typedef signed long long        Jlong;
typedef unsigned char           Jubyte;
typedef unsigned short int      Jushort;
typedef unsigned int            Juint;
typedef unsigned long long      Julong;
typedef float                   Jfloat;
typedef double                  Jdouble;
typedef char					Jboolean;
typedef char*					Jstring;	/* this is not really java-like */

/* composite value types */
typedef union _Value {
    Jboolean	b;
	Jlong		i;
	Julong		u;
	Jdouble		f;
	char*		s;
	void*		p;
	void		(*g)();
	struct {
		Jint left; Jint right;
		} 		range;
	struct {
		Juint left; Juint right;
		} 		urange;
	struct {
		Jfloat left; Jfloat right;
		} 		frange;
	struct {
		Jshort	r, g, b;
		Jshort	index;
		} 		color;
	} Value;

extern Value	comValue;

#define COM_BVALUE( v )		(comValue.b=(v), comValue)
#define COM_IVALUE( v )		(comValue.i=(v), comValue)
#define COM_UVALUE( v )		(comValue.u=(v), comValue)
#define COM_FVALUE( v )		(comValue.f=(v), comValue)
#define COM_SVALUE( v )		(comValue.s=(v), comValue)
#define COM_RVALUE( l, r )	(comValue.range.left=(l), comValue.range.right=(r), comValue)
#define COM_CVALUE(r,g,b)	(comValue.r=(r), comValue.g=(g), comValue.b=(b), comValue.index=0, comValue)

/* the following is deprecated */
#define META_BVALUE( v )	COM_BVALUE(v)
#define META_IVALUE( v )	COM_IVALUE(v)
#define META_UVALUE( v )	COM_UVALUE(v)
#define META_FVALUE( v )	COM_FVALUE(v)
#define META_SVALUE( v )	COM_SVALUE(v)
#define META_RVALUE( l, r )	COM_RVALUE((l),(r))
#define META_CVALUE(r,g,b)	COM_CVALUE(v)
							 

/* Boolean constants */
#define TRUE					1
#define FALSE					0

/* Standard COM error codes */
typedef Jint					Jerror;

/* Standard COM error codes */
#define COM_E_UNEXPECTED      0x8000ffff      /* Unexpected error */
#define COM_E_NOTIMPL         0x80004001      /* Not implemented */
#define COM_E_NOINTERFACE     0x80004002      /* Interface not supported */
#define COM_E_POINTER         0x80004003      /* Bad pointer */
#define COM_E_ABORT           0x80004004      /* Operation aborted */
#define COM_E_FAIL            0x80004005      /* Operation failed */
#define COM_E_ACCESSDENIED    0x80070005      /* Access denied */
#define COM_E_OUTOFMEMORY     0x8007000e      /* Out of memory */
#define COM_E_INVALIDARG      0x80070057      /* Invalid argument */

#define COM_SUCCEEDED(err)            ((Jerror)(err) >= 0)
#define COM_FAILED(err)               ((Jerror)(err) < 0)

#define COM_ERROR_CODE(err)           ((err) & 0xffff)
#define COM_ERROR_FACILITY(err)       (((err) >> 16) & 0x1fff)
#define COM_ERROR_SEVERITY(err)       (((err) >> 31) & 1)

#define COM_MAKE_ERROR(severity, facility, code)        \
        (((Jerror)(severity) << 31) |            \
         ((Jerror)(facility) << 16) |            \
         (Jerror)(code))

#define COM_SEVERITY_SUCCESS  0
#define COM_SEVERITY_ERROR    1

#define COM_FACILITY_NULL     0
#define COM_FACILITY_RPC      1
#define COM_FACILITY_ITF      4

/* Standard success codes */
#define COM_S_OK              0
#define COM_S_TRUE            0
#define COM_S_FALSE           1

/* Common method modifiers */
#define INLINE
#define COMCALL
#define COMDECL		Jerror		COMCALL
#define COMDECLN	Jint		COMCALL
#define COMDECLU	Juint		COMCALL
#define COMDECLF	Jfloat		COMCALL
#define COMDECLD	Jdouble		COMCALL
#define COMDECLV	void		COMCALL
#define COMDECLP	void* 		COMCALL
#define COMDECLS	char*		COMCALL
#define COMDECLB	Jboolean 	COMCALL

/* Globally Unique Identifier */
typedef struct guid {
	Juint			data1;
	Jushort			data2;
	Jushort			data3;
	Jubyte			data4[8];
	} *ComGUID, *ComIID;

#define COM_DEFINE_GUID( l,w1,w2,b1,b2,b3,b4,b5,b6,b7,b8) \
        { l, w1, w2, { b1, b2, b3, b4, b5, b6, b7, b8 } }

#define COM_ICALL( method, i )		(*(i)->vtbl->method)
#define COM_DCALL( method, d )		(*d->method)

/* COM_CALL is deprecated by COM_ICALL */
#define COM_CALL( method, o )		(*(o)->vtbl->method)

#define COM_SAME_IID( iid1, iid2 )	\
	(memcmp( iid1, iid2, sizeof(*iid1) ) == 0)

#define COM_SAME_IIDNAME( iid, name )	\
    COM_SAME_IID( iid, &comI##name##IID)

#define COM_DECLARE_CLASS( name )		\
        typedef struct _##name		name##Decl[1], *name

#define COM_ALIAS_CLASS( name, alias )		\
        typedef struct _##name		alias##Decl[1], *alias

#define COM_INHERIT_CLASS( name )		\
        struct _##name		base

#define COM_DECLARE_INTERFACE( name )	\
        typedef struct _I##name*	I##name;	\
        typedef struct _D##name*	D##name;	\
        extern struct guid comI##name##IID;		\
        struct _I##name { D##name vtbl; }

#define COM_INHERIT_INTERFACE( name )	\
        struct _D##name				base

#define COM_DECLARE_ADAM( intf, name )				\
        extern struct _D##intf		d##name;		\
        struct _I##intf		_i##name = { &d##name };

#define COM_DECLARE_ACCESSOR( name )				\
        typedef struct _D##name*	D##name;		\
        extern struct guid comI##name##IID;			\
        struct _I##name { D##name vtbl; };			\
        typedef struct { D##name vtbl; char pad[64]; }	\
            I##name##Decl[1], *I##name
		
/* The UnKnown Interface: the root of all com interfaces */

typedef void*				IGeneric;									 

#define COM_IUNKNOWN_IID COM_DEFINE_GUID(0x00000000, 0x0000, 0x0000, \
                0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x46)

#define COM_ICOMPONENT_IID COM_DEFINE_GUID(				\
        0x38b8a7d0, 0xd4ae, 0x11d3, 					\
        0xb0, 0x07, 0x00, 0xb0, 0xd0, 0x19, 0xb1, 0x55	\
        )

#define COM_IMEM_IID COM_DEFINE_GUID(					\
        0x38b8a7cf, 0xd4ae, 0x11d3, 					\
        0xb0, 0x07, 0x00, 0xb0, 0xd0, 0x19, 0xb1, 0x55	\
        )

#define COM_IITERATOR_IID COM_DEFINE_GUID(			\
    0x968062e2, 0xfd00, 0x11d3, 					\
    0xb0, 0x07, 0x00, 0xb0, 0xd0, 0x19, 0xb1, 0x55	\
	)

#define COM_ISEQUENCE_IID COM_DEFINE_GUID(0xfa2f32d7, 0x381a, 0x11d5, \
                0x8d, 0xc8, 0x00, 0x50, 0x56, 0xcb, 0x75, 0x4c)

#define COM_ISET_IID COM_DEFINE_GUID(0xfa2f32d8, 0x381a, 0x11d5, \
                0x8d, 0xc8, 0x00, 0x50, 0x56, 0xcb, 0x75, 0x4c)

#define COM_IMAP_IID COM_DEFINE_GUID(0xfa2f32f3, 0x381a, 0x11d5, \
                0x8d, 0xc8, 0x00, 0x50, 0x56, 0xcb, 0x75, 0x4c)

COM_DECLARE_INTERFACE( UnKnown );
COM_DECLARE_INTERFACE( Component );
COM_DECLARE_INTERFACE( Mem );

COM_DECLARE_ACCESSOR( Iterator );
COM_DECLARE_ACCESSOR( Sequence );
COM_DECLARE_ACCESSOR( Set );
COM_DECLARE_ACCESSOR( Map );

#define COM_INHERIT_UNKNOWN( name )							\
	COMDECL		(*query)(									\
		I##name obj, const ComIID iid, IGeneric *ihandle	\
		);													\
	COMDECLU	(*addref)( I##name obj );					\
	COMDECLU	(*release)( I##name obj )

struct _DUnKnown {
	COM_INHERIT_UNKNOWN( UnKnown );
	};

#define COM_INHERIT_COMPONENT( name )						\
	COM_INHERIT_UNKNOWN( name );							\
	COMDECLS	(*dump)( I##name ci, FILE* fp, char* bp );	\
	COMDECL		(*destruct)( I##name ci )

struct _DComponent {
	COM_INHERIT_COMPONENT( Component );
	};

#define COM_INHERIT_MEM( name )								\
    COM_INHERIT_COMPONENT( name );							\
	COMDECLP	(*alloc)( I##name mi, Juint size );			\
	COMDECLP	(*allocArray)( I##name mi, Juint size, Juint length );	\
	COMDECLP	(*allocMDArray)( I##name mi, Juint size, Jlong dim );	\
	COMDECLP	(*allocAligned)( I##name mi, Juint size, Juint align );	\
	COMDECL		(*grow)( I##name mi, Juint size );						\
	COMDECLV	(*free)( I##name mi, void *ptr );						\
	COMDECLN	(*avail)( I##name mi )

struct _DMem {
	COM_INHERIT_MEM( Mem );
	};

#define COM_INHERIT_ITERATOR( name )							\
    COM_INHERIT_UNKNOWN( name );								\
	COMDECL		(*reset)( I##name it );							\
	COMDECLB	(*hasMoreElements)( I##name it );				\
	COMDECLP	(*nextElement)(  I##name it )				

struct _DIterator {
	COM_INHERIT_ITERATOR( Iterator );
	};

struct _DSequence {
	COM_INHERIT_ITERATOR( Sequence );
	
	COMDECL		(*pushElement)( ISequence it, IMem mem, IGeneric data );	
	COMDECLP	(*popElement)( ISequence it );							
	COMDECL		(*appendElement)( ISequence it, IMem mem, IGeneric data );
	COMDECL		(*removeElement)( ISequence it, IGeneric data );			
	COMDECL		(*removeAllElements)( ISequence it );
	};

struct _DSet {
	COM_INHERIT_ITERATOR( Set );

	COMDECL		(*makeEmpty)( ISet si );
	COMDECL		(*makeUniv)( ISet si );
	COMDECL		(*copy)( ISet si, ISet result ); 
	COMDECL		(*unite)( ISet si, ISet another, ISet result ); 
	COMDECL		(*intersect)( ISet si, ISet another, ISet result ); 
	COMDECL		(*complement)( ISet si, ISet result ); 
	COMDECL		(*equals)( ISet si, ISet another ); 
	COMDECL		(*isEmpty)( ISet si ); 
	COMDECL		(*isMember)( ISet si, Jint m ); 
	COMDECL		(*includeMember)( ISet si, Jint m ); 
	COMDECL		(*excludeMember)( ISet si, Jint m ); 
	};

struct _DMap {
	COM_INHERIT_ITERATOR( Map );

	COMDECL		(*put)( IMap it, IGeneric key, IGeneric data );
	COMDECLP	(*get)( IMap it, IGeneric key );
	COMDECLP	(*key)( IMap it );
	};

/* now these macros are deprecated */
#define COM_DECLARE_ITERATOR( name )		IIteratorDecl	name
#define COM_DECLARE_SEQUENCE( name )		ISequenceDecl	name
#define COM_DECLARE_SET( name )				ISetDecl		name
#define COM_DECLARE_MAP( name )				IMapDecl		name
#define ACCESSOR_SIZE						sizeof(IIteratorDecl)

#define comQuery( o, iid, i )       \
    COM_ICALL( query, (IUnKnown)(o) )( (IUnKnown)(o), iid, i )

#define comAddRef( o )              \
    COM_ICALL( addref, (IUnKnown)(o) )( (IUnKnown)o )

#define comRelease( o )             \
    COM_ICALL( release, (IUnKnown)(o) )( (IUnKnown)o )

#define comRef( o )                 \
    (COM_ICALL( addref, (IUnKnown)(o) )( (IUnKnown)o ), (o) )

#define comInstanceOf( obj, name )	\
    (comQuery( (obj), &comI##name##IID, 0 ) != COM_E_NOINTERFACE)

#define comDump( obj, fp, bp )      \
    (COM_ICALL( dump, (IComponent)(obj) )( (IComponent)(obj), fp, bp ))

#define comDestruct( obj )	\
    (COM_ICALL( destruct, (IComponent)(obj) )( (IComponent)(obj) ))

#define comAlloc( obj, size )    \
    (COM_ICALL( alloc, (IMem)(obj) )( (IMem)(obj), size ))

#define comAllocArray( obj, size, len )          \
    (COM_ICALL( allocArray, (IMem)(obj) )( (IMem)(obj), size, len ))

#define comAllocMDArray( obj, size, dim )        \
    (COM_ICALL( allocArray, (IMem)(obj) )( (IMem)(obj), size, dim ))

#define comAllocAligned( obj, size, align )      \
    (COM_ICALL( allocArray, (IMem)(obj) )( (IMem)(obj), size, align ))

#define comGrow( obj, size )     \
    (COM_ICALL( grow, (IMem)(obj) )( (IMem)(obj), size ))

#define comFree( obj, ptr )      \
    (COM_ICALL( free, (IMem)(obj) )( (IMem)(obj), ptr ))

#define comAvail( obj )          \
    (COM_ICALL( avail, (IMem)(obj) )( (IMem)(obj) ))

#define comReset( it )				\
    (COM_ICALL( reset, it )( it ))

#define comHasMoreElements( it )	\
    (COM_ICALL( hasMoreElements, it )( it ))

#define comNextElement( it )		\
    (COM_ICALL( nextElement, it )( it ))

#define comSeqPushElement( it, mem, data )	\
    (COM_ICALL( pushElement, it )( it, mem, data ))

#define comSeqPopElement( it )			\
    (COM_ICALL( popElement, it )( it ))

#define comSeqAppendElement( it, mem, data )	\
    (COM_ICALL( appendElement, it )( it, mem, data ))

#define comSeqRemoveElement( it, data )		\
    (COM_ICALL( removeElement, it )( it, data ))

#define comSeqRemoveAllElements( it )		\
    (COM_ICALL( removeAllElements, it )( it ))

#define comSetMakeEmpty( it )		\
    (COM_ICALL( makeEmpty, it )( it ))

#define comSetMakeUniv( it )		\
    (COM_ICALL( makeUnit, it )( it ))

#define comSetCopy( it, result )			\
    (COM_ICALL( copy, it )( it, result ))

#define comSetUnite( it, another, result )	\
    (COM_ICALL( unite, it )( it, another, result ))

#define comSetIntersect( it, another, result )	\
    (COM_ICALL( intersect, it )( it, another, result ))

#define comSetComplement( it, result )	\
    (COM_ICALL( complement, it )( it, result ))

#define comSetEquals( it, another )		\
    (COM_ICALL( equals, it )( it, another ))

#define comSetIsEmpty( it )	\
    (COM_ICALL( isEmpty, it )( it ))

#define comSetIncludeMember( it, i )	\
    (COM_ICALL( includeMember, it )( it, i ))

#define comSetExcludeMember( it, i )	\
    (COM_ICALL( excludeMember, it )( it, i ))

#define comMapPut( it, key, data )		\
    (COM_ICALL( put, it )( it, key, data ))

#define comMapGet( it, key )			\
    (COM_ICALL( get, it )( it, key ))

#define comMapKey( it )			\
    (COM_ICALL( key, it )( it ))

extern IMem		iheap;

#ifdef __cplusplus

class CUnKnown {
  public:
	IUnKnown			intf;
	
	CUnKnown( void ) { intf = 0; }
	CUnKnown( IGeneric i ) {
		intf = (IUnKnown)i; COM_ICALL( addref,  (IUnKnown)i )(  (IUnKnown)i );
		}
	CUnKnown( CUnKnown& another ) {
		set( another.intf );
		}
	~CUnKnown( void ) { COM_ICALL( release, intf )( intf ); }
	CUnKnown& operator=( CUnKnown another ) {
		set( another.intf );
		return *this;
		}
	CUnKnown& operator=( IGeneric intf ) {
		set( intf );
		return *this;
		}
	inline Jerror query( const ComIID iid, CUnKnown handle ) {
		IUnKnown	i;
		Jerror		rc;
		
		rc = COM_ICALL( query, intf )( intf, iid, (IGeneric *)&i );
		if( handle.intf != i ) {
			COM_ICALL( release, handle.intf );
			COM_ICALL( addref, i )( i );
			handle.intf = i;
			}
		return rc;
		}
	inline void set( IGeneric another ) {
		if( intf == (IUnKnown)another ) return;
		if( intf )
			COM_ICALL( release, intf )( intf );
		intf = (IUnKnown)another;
		if( another )
			COM_ICALL( addref, (IUnKnown)another )( (IUnKnown)another );
		return;
		}
	};

/*
 * The circular dependence prevents
 * being the subclass of CComponent
 */
class CMem : public CUnKnown {
  public:
	CMem( IGeneric i ) : CUnKnown( i ) {
		assert( comInstanceOf( i, Component ) );
		assert( comInstanceOf( i, Mem ) );
		}
	CMem( void ) {}
	
	inline char* 	dump( FILE* fp, char* bp ) {
		return COM_ICALL( dump, (IMem)intf )(
			(IMem)intf, fp, bp
			);
		}
	inline Jerror destruct( void ) {
		return COM_ICALL( destruct, (IMem)intf )( (IMem)intf );
		}
	
	inline void* alloc( Juint size ) {
		return COM_ICALL( alloc, (IMem)intf )( (IMem)intf, size );
		}
	inline void* allocArray( Juint size, Juint length ) {
		return COM_ICALL( allocArray, (IMem)intf )( (IMem)intf, size, length );
		}
	inline void* allocMDArray( Juint size, Jlong dim ) {
		return COM_ICALL( allocMDArray, (IMem)intf )( (IMem)intf, size, dim );
		}
	inline void* allocAligned( Juint size, Juint align ) {
		return COM_ICALL( allocMDArray, (IMem)intf )( (IMem)intf, size, align );
		}
	inline Jerror grow( Juint size ) {
		return COM_ICALL( grow, (IMem)intf )( (IMem)intf, size );
		}
	inline void free( void* ptr ) {
		return COM_ICALL( free, (IMem)intf )( (IMem)intf, ptr );
		}
	inline Jint avail( void ) {
		return COM_ICALL( avail, (IMem)intf )( (IMem)intf );
		}
	};

class CComponent : public CUnKnown {
  public:
	CComponent( IGeneric i ) : CUnKnown( i ) {
		assert( comInstanceOf( i, Component ) );
		}
	CComponent( void ) {}
	CComponent( CComponent& another ) {
		set( another.intf );
		}

	inline char* 	dump( FILE* fp, char* bp ) {
		return COM_ICALL( dump, (IComponent)intf )(
			(IComponent)intf, fp, bp
			);
		}
	inline Jerror	destruct( void ) {
		return COM_ICALL( destruct, (IComponent)intf )( (IComponent)intf );
		}
	};

class CIterator : public CUnKnown {
  public:
	
	CIterator( IGeneric i ) : CUnKnown( i ) {}
	CIterator( void ) {}
	
	inline Jerror		reset( void ) {
		return COM_CALL( reset, (IIterator)intf )( (IIterator)intf );
		} 
	inline Jboolean		hasMoreElements( void ) {
		return COM_CALL( hasMoreElements, (IIterator)intf )( (IIterator)intf );
		} 
	inline IGeneric		nextElement( void ) {
		return COM_CALL( nextElement, (IIterator)intf )( (IIterator)intf );
		}
	};

class CSequence : public CIterator {
  public:
	CSequence( IGeneric i ) : CIterator( i ) {}
	CSequence( void ) {}
	
	inline Jerror		pushElement( CMem mem, IGeneric data ) {
		return COM_CALL( pushElement, (ISequence)intf )(
			(ISequence)intf, (IMem)mem.intf, data
			);
		} 
	inline IGeneric		popElement( void ) {
		return COM_CALL( popElement, (ISequence)intf )( (ISequence)intf );
		} 
	inline Jerror		appendElement( CMem mem, IGeneric data ) {
		return COM_CALL( appendElement, (ISequence)intf )(
			(ISequence)intf, (IMem)mem.intf, data
			);
		}
	inline Jerror		removeElement( IGeneric data ) {
		return COM_CALL( removeElement, (ISequence)intf )(
			(ISequence)intf, data
			);
		}
	inline Jerror		removeAllElements( void ) {
		return COM_CALL( removeAllElements, (ISequence)intf )(
			(ISequence)intf
			);
		}
	};

class CSet : public CIterator {
  public:
	CSet( IGeneric i ) : CIterator( i ) {}
	CSet( void ) {}

	inline Jerror		makeEmpty( void ) {
		return COM_ICALL( makeEmpty, (ISet)intf )( (ISet)intf );
		}
	inline Jerror		makeUniv( void ) {
		return COM_ICALL( makeEmpty, (ISet)intf )( (ISet)intf );
		}
	inline Jerror		copy( CSet result ) {
		return COM_ICALL( copy, (ISet)intf )( (ISet)intf, (ISet)result.intf );
		}
	inline Jerror		unite( CSet another, CSet result ) {
		return COM_ICALL( unite, (ISet)intf )(
			(ISet)intf, (ISet)another.intf, (ISet)result.intf
			);
		}
	inline Jerror		intersect( CSet another, CSet result ) {
		return COM_ICALL( intersect, (ISet)intf )(
			(ISet)intf, (ISet)another.intf, (ISet)result.intf
			);
		}
	inline Jerror		complement( CSet result ) {
		return COM_ICALL( complement, (ISet)intf )(
			(ISet)intf, (ISet)result.intf
			);
		}
	inline Jboolean		equals( CSet another ) {	
		return COM_ICALL( equals, (ISet)intf )(
			(ISet)intf, (ISet)another.intf
			);
		}
	inline Jboolean    isEmpty( void ) {
		return COM_ICALL( isEmpty, (ISet)intf )(
			(ISet)intf
			);
		}
	inline Jboolean		isMember( Jint m ) {
		return COM_ICALL( isMember, (ISet)intf )(
			(ISet)intf, m
			);
		}
	inline Jerror		includeMember( Jint m ) {
		return COM_ICALL( includeMember, (ISet)intf )(
			(ISet)intf, m
			);
		}
	inline Jerror		excludeMember( Jint m ) {
		return COM_ICALL( excludeMember, (ISet)intf )(
			(ISet)intf, m
			);
		}
	};
#endif

#endif



